from mailer import Mailer, Message

from .. import root, config, logger

LOGIN_HTML = root / "html" / "login.html"


# pylint: disable=too-few-public-methods


class EmailMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def send_auth_email(email, auth_url):
        logger.info(f"Login here to use Spotify API: {auth_url}")
        mailer = Mailer(**config.email)
        # pylint: disable=no-member
        html_content = LOGIN_HTML.read_text().replace(
            "SPOTIFY_AUTHENTICATION_URL", auth_url
        )
        message = Message(From=config.email.usr or email, To=email, charset="utf-8")
        message.Subject = "Spotify API Authentication"
        message.Html = html_content
        message.Body = f"Login here to use Spotify API: {auth_url}"
        mailer.send(message)



import time
import random
from datetime import date, timedelta
from itertools import chain

from pony.orm import get
from pony.orm.core import CacheIndexError

from .. import logger
from ..util import normalize_features
from ..cache import Genre, Artist, Playlist, select, db_session
from ..constants import TimeRange


class RecommenderMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @db_session
    def fetch_playlists(self):
        fetched_ids = set(select(p.id for p in Playlist))
        results = chain(
            self.user_playlists("particledetector").all(),
            self.user_playlists("thesoundsofspotify").all(),
        )
        for playlist in results:
            logger.info(f"Fetching {playlist.name}")
            if playlist.id not in fetched_ids:
                Playlist.from_dict(playlist)
            fetched_ids.add(playlist.id)

    @db_session
    def fetch_user_top(self, time_range):
        self.user.top_artists.clear()
        for artist in self.current_user_top_artists(
            limit=50, time_range=time_range
        ).iterall():
            if self.is_disliked_artist(artist):
                continue

            if Artist.exists(id=artist.id):
                self.user.top_artists.add(Artist[artist.id])
            else:
                try:
                    artist = Artist.from_dict(artist)
                except CacheIndexError:
                    artist = Artist[artist.id]
                self.user.top_artists.add(artist)
        self.user.top_genres = self.user.top_artists.genres.distinct().keys() - self.user.disliked_genres
        if self.user.top_expires_at is None:
            self.user.top_expires_at = {}
        self.user.top_expires_at[TimeRange(time_range).value] = time.mktime(
            (date.today() + timedelta(days=1)).timetuple()
        )

    @db_session
    def genre_playlist(
        self, genre, popularity=Playlist.Popularity.SOUND
    ):  # pylint: disable=no-self-use
        if not Playlist.exists():
            raise Exception('You have to call "fetch_playlists" first.')

        return Playlist.get(
            genre=genre, popularity=Playlist.Popularity(popularity).value
        )

    @db_session
    def top_artists(self, time_range=TimeRange.SHORT_TERM):
        if self.user.top_expired(time_range):
            self.fetch_user_top(time_range)
        return self.user.top_artists

    @db_session
    def top_genres(self, time_range=TimeRange.SHORT_TERM):
        if self.user.top_expired(time_range):
            self.fetch_user_top(time_range)
        return self.user.top_genres

    def order_by(self, features, tracks):
        audio_features = self.audio_features(tracks=tracks)
        track_ids = [a.id for a in audio_features]
        audio_features = [a.to_dict(list(features.keys())) for a in audio_features]
        audio_features = normalize_features(audio_features, track_ids)
        for feature, direction in features.items():
            audio_features[feature] *= direction
        audio_features["total"] = audio_features.sum(axis=1)
        return audio_features.sort_values("total").index.tolist()

    def fill_with_related_artists(self, artists, limit=5):
        tries = 5
        if len(artists) >= limit:
            return artists

        artist_set = set(artists)
        artist_list = list(artists)
        while len(artist_set) < limit and tries:
            tries -= 1
            related_artists = {
                a.id
                for a in self.artist_related_artists(random.choice(artist_list))
                if self.is_not_disliked_artist(a)
            }
            related_artists_limit = min(
                random.randint(1, limit - len(artist_set)), len(related_artists)
            )
            artist_set |= set(random.sample(related_artists, related_artists_limit))
        return artist_set

    @db_session
    def recommend_by_top_artists(
        self,
        artist_limit=2,
        track_limit=100,
        use_related=True,
        features_order=None,
        time_range=TimeRange.SHORT_TERM,
        **kwargs,
    ):
        """Get a list of recommended songs.

        Returns:
            list: List of tracks
        """
        artists = self.top_artists(
            time_range=time_range
        ).select().without_distinct().random(
            artist_limit
        )
        if use_related:
            artists = self.fill_with_related_artists([a.id for a in artists])
        tracks = self.recommendations(seed_artists=artists, limit=track_limit, **kwargs)
        tracks = list(filter(self.is_not_disliked_track, tracks))
        if features_order:
            tracks = self.order_by(features_order, tracks)
        return tracks

    @db_session
    def is_disliked_artist(self, artist):
        return (
            artist.id in set(self.user.disliked_artists.id.distinct().keys())
            or bool(set(artist.genres or []) & set(self.user.disliked_genres))
        )

    @db_session
    def is_not_disliked_artist(self, artist):
        return not self.is_disliked_artist(artist)

    @db_session
    def is_not_disliked_track(self, track):
        return all(self.is_not_disliked_artist(a) for a in track.artists)

    @db_session
    def disliked_artists(self):
        return list(self.user.disliked_artists)

    @db_session
    def disliked_genres(self):
        return list(self.user.disliked_genres)

    @db_session
    def dislike_artist(self, artist):
        if isinstance(artist, str):
            artist = get(a for a in Artist if a.id == artist or a.name == artist)
        elif isinstance(artist, dict):
            artist = Artist.get(artist.id) or Artist.from_dict(artist)
        self.user.dislike(artist=artist)

    @db_session
    def dislike_genre(self, genre):
        if isinstance(genre, str):
            genre = genre.lower()
            genre = Genre.get(name=genre) or Genre(name=genre)
        self.user.dislike(genre=genre)



import uuid
import socket
import threading
from pathlib import Path
from wsgiref.simple_server import make_server

import hug
from pony.orm import get
from cachecontrol import CacheControlAdapter
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from cachecontrol.caches.file_cache import FileCache

from .. import root, config, logger
from ..cache import User, select, db_session
from ..constants import API, AuthFlow, AllScopes
from ..exceptions import SpotifyCredentialsException

AUTH_HTML_FILE = root / "html" / "auth_message.html"
CACHE_FILE = Path.home() / ".cache" / "spfy" / ".web_cache"


class AuthMixin:

    def __init__(
        self,
        *args,
        client_id=None,
        client_secret=None,
        redirect_uri=None,
        user_id=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.client_id = client_id or config.app.client_id
        self.client_secret = client_secret or config.app.client_secret
        self.redirect_uri = self._get_redirect_uri(redirect_uri)
        self.session = None
        self.user_id = user_id
        self.callback_reached = threading.Event()

    @staticmethod
    def _get_redirect_uri(redirect_uri):
        redirect_uri = (
            redirect_uri
            or config.app.redirect_uri
            or f"http://{socket.gethostname()}.local"
        )
        if config.auth.callback.enabled and config.auth.callback.port and redirect_uri:
            redirect_uri += f":{config.auth.callback.port}"
        return redirect_uri

    @staticmethod
    def get_session(*args, **kwargs):
        session = OAuth2Session(*args, **kwargs)
        cache_adapter = CacheControlAdapter(
            cache=FileCache(CACHE_FILE),
            pool_connections=config.http.connections,
            pool_maxsize=config.http.connections,
            max_retries=config.http.retries,
        )
        session.mount("http://", cache_adapter)
        return session

    @property
    @db_session
    def user(self):
        return User[self.user_id]

    @property
    def is_authenticated(self):
        return bool(self.session and self.session.authorized)

    @db_session
    def authenticate_user(
        self,
        username=None,
        email=None,
        code=None,
        state=None,
        auth_response=None,
        scope=AllScopes,
    ):
        session = self.session or self.get_session(
            self.client_id,
            redirect_uri=self.redirect_uri,
            scope=scope,
            auto_refresh_url=API.TOKEN.value,
        )
        if self.user_id:
            user = User.get(id=self.user_id)
            if user and user.token:
                session.token = user.token
                session.token_updater = User.token_updater(user.id)
                return session

        if username or email:
            user = get(u for u in User if u.username == username or u.email == email)
            if user:
                self.user_id = user.id
                session.token = user.token
                session.token_updater = User.token_updater(user.id)
                return session

        if code or auth_response:
            token = session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
                code=code,
                state=state,
                authorization_response=auth_response,
            )
            user_details = self.current_user()
            user = select(
                u
                for u in User
                if u.username == user_details.id or u.email == user_details.email
            ).for_update().get()
            if user:
                user.token = token
                if user.id != self.user_id:
                    self.user_id = user.id
            elif self.user_id:
                user = User.get_for_update(id=self.user_id)
                if user:
                    user.token = token
            if not user:
                self.user_id = self.user_id or uuid.uuid4()
                user_details["user_id"] = self.user_id
                user_details["token"] = token
                user = User.from_dict(user_details)
            return session

        return session

    @db_session
    def authenticate_server(self):
        default_user = User.default()
        self.user_id = default_user.id
        session = self.session or self.get_session(
            client=BackendApplicationClient(self.client_id)
        )
        session.token_updater = User.token_updater(default_user.id)
        if default_user.token:
            session.token = default_user.token
        else:
            default_user.token = session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
        return session

    def authenticate(self, flow=config.auth.flow, **auth_params):
        if not (self.client_id and self.client_secret):
            raise SpotifyCredentialsException

        flow = AuthFlow(flow)
        if flow == AuthFlow.CLIENT_CREDENTIALS:
            self.session = self.authenticate_server()
        elif flow == AuthFlow.AUTHORIZATION_CODE:
            self.session = self.authenticate_user(**auth_params)
            if not self.session.token:
                if config.auth.callback.enabled:
                    self.start_callback()
                authorization_url, _ = self.session.authorization_url(
                    API.AUTHORIZE.value
                )
                if config.auth.send_auth_url_to_email:
                    email = auth_params.get("email") or config.auth.email
                    self.send_auth_email(email, authorization_url)
                else:
                    print(f"Login here: {authorization_url}")
                self.wait_for_authorization()

    def wait_for_authorization(self):
        if not config.auth.callback.enabled:
            url = input("Paste the URL you are redirected to:")
            self.session = self.authenticate_user(auth_response=url)
        else:
            self.callback_reached.wait(config.auth.callback.timeout)
            self.stop_callback()

    def stop_callback(self):
        if self.httpd:
            self.httpd.shutdown()

    def start_callback(self):
        self.callback_reached.clear()

        # pylint: disable=unused-variable

        @hug.get("/", output=hug.output_format.html)
        def callback(code: hug.types.text, state: hug.types.text):
            html = AUTH_HTML_FILE.read_text()  # pylint: disable=no-member
            try:
                self.session = self.authenticate_user(code=code, state=state)
                html = html.replace("SPOTIFY_AUTH_MESSAGE", "Successfully logged in!")
                html = html.replace("BACKGROUND_COLOR", "#65D46E")
            except Exception as exc:
                logger.exception(exc)
                html = html.replace(
                    "SPOTIFY_AUTH_MESSAGE", "Could not get authorization token."
                )
                html = html.replace("BACKGROUND_COLOR", "#EC2E50")
            finally:
                self.callback_reached.set()
            return html

        api = __hug__.http.server(None)  # pylint: disable=undefined-variable
        self.httpd = make_server("", config.auth.callback.port, api)
        threading.Thread(target=self.httpd.serve_forever, daemon=True).start()



import os
import pathlib

from setuptools import setup, find_packages

CONFIGDIR = pathlib.Path.home() / '.config' / 'spfy'
CONFIGDIR.mkdir(parents=True, exist_ok=True)
with open('spfy/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break

    else:
        version = '0.0.1'
with open('README.md', 'rb') as f:
    readme = f.read().decode('utf-8')
REQUIRES = [
    'addict',
    'backoff',
    'colorthief',
    'fire',
    'first',
    'gunicorn',
    'hug',
    'oauthlib',
    'requests',
    'requests_oauthlib',
    'mailer',
    'pony',
    'cached_property',
    'kick',
    'cachecontrol',
    'lockfile',
    'python-unsplash-async',
    'aiohttp',
    'psycopg2-binary',
    'pycountry',
    'pandas',
    'ujson',
    'aioredis',
    'msgpack',
]
try:
    if os.uname().sysname == 'Linux':
        REQUIRES.append('pyalsaaudio')
except:
    pass
setup(
    name='spfy',
    version=version,
    description='',
    long_description=readme,
    author='Alin Panaitiu',
    author_email='alin.p32@gmail.com',
    maintainer='Alin Panaitiu',
    maintainer_email='alin.p32@gmail.com',
    url='https://github.com/alin23/spfy',
    license='MIT/Apache-2.0',
    keywords=['spotify', 'spfy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    install_requires=REQUIRES,
    tests_require=['coverage', 'pytest', 'tox', 'pylint', 'mypy'],
    packages=find_packages(),
    package_data={'spfy': ['config/*.toml', 'html/*.html']},
    data_files=[(str(CONFIGDIR), ['spfy/config/config.toml'])],
    entry_points={
        'console_scripts': [
            'spotify = spfy.wrapper:main', 'spotify_async = spfy.asynch.wrapper:main'
        ]
    },
)



import re
from enum import Enum

from .frange import frange


class ItemType(Enum):
    TRACKS = "tracks"
    TRACK = "track"
    PLAYLIST = "playlist"
    ARTIST = "artist"
    ALBUM = "album"


class TimeRange(Enum):
    SHORT_TERM = "short_term"
    MEDIUM_TERM = "medium_term"
    LONG_TERM = "long_term"


class VolumeBackend(Enum):
    SPOTIFY = "spotify"
    APPLESCRIPT = "applescript"
    ALSA = "alsa"
    LINUX = "linux"


class AuthFlow(Enum):
    AUTHORIZATION_CODE = "authorization_code"
    CLIENT_CREDENTIALS = "client_credentials"


class AudioFeature(Enum):
    ACOUSTICNESS = "acousticness"
    DANCEABILITY = "danceability"
    DURATION_MS = "duration_ms"
    ENERGY = "energy"
    INSTRUMENTALNESS = "instrumentalness"
    KEY = "key"
    LIVENESS = "liveness"
    LOUDNESS = "loudness"
    MODE = "mode"
    POPULARITY = "popularity"
    SPEECHINESS = "speechiness"
    TEMPO = "tempo"
    TIME_SIGNATURE = "time_signature"
    VALENCE = "valence"


class AudioFeatureRange(Enum):
    ACOUSTICNESS = frange(0.0, 1.01, 0.01)
    DANCEABILITY = frange(0.0, 1.01, 0.01)
    DURATION_MS = range((30 * 60 * 1000) + 1)
    ENERGY = frange(0.0, 1.01, 0.01)
    INSTRUMENTALNESS = frange(0.0, 1.01, 0.01)
    KEY = range(0, 12)
    LIVENESS = frange(0.0, 1.01, 0.01)
    LOUDNESS = frange(-60.0, 0.1, 0.1)
    MODE = range(2)
    POPULARITY = range(101)
    SPEECHINESS = frange(0.0, 1.01, 0.01)
    TEMPO = range(321)
    TIME_SIGNATURE = range(9)
    VALENCE = frange(0.0, 1.01, 0.01)


class Scope(Enum):
    PLAYLIST_READ_PRIVATE = "playlist-read-private"
    PLAYLIST_READ_COLLABORATIVE = "playlist-read-collaborative"
    PLAYLIST_MODIFY_PUBLIC = "playlist-modify-public"
    PLAYLIST_MODIFY_PRIVATE = "playlist-modify-private"
    STREAMING = "streaming"
    UGC_IMAGE_UPLOAD = "ugc-image-upload"
    USER_FOLLOW_MODIFY = "user-follow-modify"
    USER_FOLLOW_READ = "user-follow-read"
    USER_LIBRARY_READ = "user-library-read"
    USER_LIBRARY_MODIFY = "user-library-modify"
    USER_READ_PRIVATE = "user-read-private"
    USER_READ_BIRTHDATE = "user-read-birthdate"
    USER_READ_EMAIL = "user-read-email"
    USER_TOP_READ = "user-top-read"
    USER_READ_PLAYBACK_STATE = "user-read-playback-state"
    USER_MODIFY_PLAYBACK_STATE = "user-modify-playback-state"
    USER_READ_CURRENTLY_PLAYING = "user-read-currently-playing"
    USER_READ_RECENTLY_PLAYED = "user-read-recently-played"


AllScopes = [scope.value for scope in Scope]


class API(Enum):
    PREFIX = "https://api.spotify.com"
    ALBUM = "/v1/albums/{id}"
    ALBUM_TRACKS = "/v1/albums/{id}/tracks"
    ALBUMS = "/v1/albums"
    ARTIST = "/v1/artists/{id}"
    ARTIST_ALBUMS = "/v1/artists/{id}/albums"
    ARTIST_RELATED_ARTISTS = "/v1/artists/{id}/related-artists"
    ARTIST_TOP_TRACKS = "/v1/artists/{id}/top-tracks"
    ARTISTS = "/v1/artists"
    AUDIO_ANALYSIS = "/v1/audio-analysis/{id}"
    AUDIO_FEATURES_SINGLE = "/v1/audio-features/{id}"
    AUDIO_FEATURES_MULTIPLE = "/v1/audio-features"
    RECOMMENDATIONS = "/v1/recommendations"
    RECOMMENDATIONS_GENRES = "/v1/recommendations/available-genre-seeds"
    AUTHORIZE = "https://accounts.spotify.com/authorize"
    TOKEN = "https://accounts.spotify.com/api/token"
    CATEGORIES = "/v1/browse/categories"
    CATEGORY = "/v1/browse/categories/{id}"
    CATEGORY_PLAYLISTS = "/v1/browse/categories/{id}/playlists"
    FEATURED_PLAYLISTS = "/v1/browse/featured-playlists"
    NEW_RELEASES = "/v1/browse/new-releases"
    ME = "/v1/me"
    MY_ALBUMS = "/v1/me/albums"
    MY_ALBUMS_CONTAINS = "/v1/me/albums/contains"
    MY_FOLLOWING = "/v1/me/following"
    MY_FOLLOWING_CONTAINS = "/v1/me/following/contains"
    MY_PLAYLISTS = "/v1/me/playlists"
    MY_TOP = "/v1/me/top/{type}"
    MY_TRACKS = "/v1/me/tracks"
    MY_TRACKS_CONTAINS = "/v1/me/tracks/contains"
    PLAYLIST = "/v1/users/{user_id}/playlists/{playlist_id}"
    PLAYLISTS = "/v1/users/{user_id}/playlists"
    PLAYLIST_FOLLOWERS = "/v1/users/{owner_id}/playlists/{playlist_id}/followers"
    PLAYLIST_FOLLOWERS_CONTAINS = "/v1/users/{user_id}/playlists/{playlist_id}/followers/contains"
    PLAYLIST_IMAGES = "/v1/users/{user_id}/playlists/{playlist_id}/images"
    PLAYLIST_TRACKS = "/v1/users/{user_id}/playlists/{playlist_id}/tracks"
    SEARCH_ALBUM = "/v1/search?type=album"
    SEARCH_ARTIST = "/v1/search?type=artist"
    SEARCH_PLAYLIST = "/v1/search?type=playlist"
    SEARCH_TRACK = "/v1/search?type=track"
    TRACK = "/v1/tracks/{id}"
    TRACKS = "/v1/tracks"
    USER = "/v1/users/{user_id}"
    CURRENTLY_PLAYING = "/v1/me/player/currently-playing"
    DEVICES = "/v1/me/player/devices"
    NEXT = "/v1/me/player/next"
    PAUSE = "/v1/me/player/pause"
    PLAY = "/v1/me/player/play"
    PLAYER = "/v1/me/player"
    PREVIOUS = "/v1/me/player/previous"
    RECENTLY_PLAYED = "/v1/me/player/recently-played"
    REPEAT = "/v1/me/player/repeat"
    SEEK = "/v1/me/player/seek"
    SHUFFLE = "/v1/me/player/shuffle"
    VOLUME = "/v1/me/player/volume"


VOLUME_FADE_SECONDS = 5 * 60
DEVICE_ID_RE = re.compile(r"[a-zA-Z0-9]{40}")
PLAYLIST_URI_RE = re.compile(r"spotify:user:[^:]+:playlist:[^:]+")



#!/usr/bin/env python3
import fire
import kick
from pony.orm import db_session

from . import APP_NAME, config, logger
from .client import SpotifyClient
from .mixins import PlayerMixin, RecommenderMixin


class Spotify(SpotifyClient, PlayerMixin, RecommenderMixin):
    """Spotify high-level wrapper."""
    cli = False

    def __init__(self, *args, email=None, username=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.email = email or config.auth.email
        self.username = username or config.auth.username

    def auth(self, email=config.auth.email, username=config.auth.username):
        if self.cli and not self.is_authenticated:
            try:
                self.authenticate(email=email, username=username)
            except Exception as exc:
                logger.exception(exc)
        return self

    def __dir__(self):
        names = super().__dir__()
        names = [name for name in names if not name.startswith("_") and name != "user"]
        return names

    @staticmethod
    def update_config(name="config"):
        kick.update_config(APP_NAME.lower(), variant=name)


def main():
    """Main function."""
    try:
        Spotify.cli = True
        with db_session:
            fire.Fire(Spotify)
    except KeyboardInterrupt:
        print("Quitting")


if __name__ == "__main__":
    main()
